/*
** Copyright 2002  Alexandre Duret-Lutz <duret_g@epita.fr>
**
** This file is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published
** by the Free Software Foundation; either version 2 of the License,
** or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston,
** MA 02111-1307 USA
*/

/*-----------------------------------------------------------.
| DO NOT MODIFY.  This file contains portions extracted from |
| several source files.  Change these files instead.         |
`-----------------------------------------------------------*/

#ifndef HEROES__PLUGIN__H
#define HEROES__PLUGIN__H

#include "pconf.h"

/*****************************************************************
******************************************************************
*** extracted from ./system.h
**/

#include <stdio.h>
#include <sys/types.h>

#if STDC_HEADERS
# include <stdlib.h>
#endif

#include <ctype.h>
#include <sys/stat.h>
#if HAVE_FCNTL_H
# include <fcntl.h>
#else
# include <sys/file.h>
#endif
#include <assert.h>
#include <math.h>
#include <errno.h>

#if HAVE_STRING_H
# if HAVE_MEMORY_H && ! STDC_HEADERS
#  include <memory.h>
# endif
# include <string.h>
#else
# if HAVE_STRINGS_H
#  include <strings.h>
# endif
#endif

#if HAVE_UNISTD_H
# include <unistd.h>
#endif

#if HAVE_DIRENT_H
# include <dirent.h>
#else
# define dirent direct
# if HAVE_SYS_NDIR_H
#  include <sys/ndir.h>
# endif
# if HAVE_SYS_DIR_H
#  include <sys/dir.h>
# endif
# if HAVE_NDIR_H
#  include <ndir.h>
# endif
#endif

#if HAVE_DIRECT_H
# include <direct.h>
#endif

#if HAVE_IO_H
# include <io.h>
#endif

#if HAVE_VPRINTF || HAVE_DOPRNT
# if STDC_HEADERS
#  include <stdarg.h>
#  define VA_START(args, lastarg) va_start(args, lastarg)
# else
#  include <varargs.h>
#  define VA_START(args, lastarg) va_start(args)
# endif
#else
# define va_alist a1, a2, a3, a4, a5, a6, a7, a8
# define va_dcl char *a1, *a2, *a3, *a4, *a5, *a6, *a7, *a8;
#endif

#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#if HAVE_MMAP
# include <sys/mman.h>
#endif

/*****************************************************************
******************************************************************
*** extracted from ./system.h
**/

/* Take care of NLS matters.  */

#if HAVE_LOCALE_H
# include <locale.h>
#endif
#if !HAVE_SETLOCALE
# define setlocale(Category, Locale) /* empty */
#endif

#if ENABLE_NLS
# include <libintl.h>
# define _(Text) gettext (Text)
#else
# undef bindtextdomain
# define bindtextdomain(Domain, Directory) /* empty */
# undef textdomain
# define textdomain(Domain) /* empty */
# define _(Text) Text
# define ngettext(msg1, msg2, n) ((n) == 1 ? (msg1) : (msg2))
#endif
#define N_(Text) Text

/* boolean type */

#if HAVE_STDBOOL_H
# include <stdbool.h>
#else
typedef enum {false = 0, true = 1} bool;
#endif

#ifndef M_PI
# define M_PI	3.14159265358979323846  /* pi */
#endif
#ifndef M_PI_2
# define M_PI_2	1.57079632679489661923  /* pi/2 */
#endif
#ifndef M_PI_4
# define M_PI_4 0.78539816339744830962  /* pi/4 */
#endif

/* common integer sizes */

#if HAVE_STDINT_H
# include <stdint.h>
typedef uint32_t	a_u32;
typedef uint16_t	a_u16;
typedef uint8_t		a_u8;
typedef int32_t		a_s32;
typedef int16_t		a_s16;
typedef int8_t		a_s8;
#define U16_MAX		UINT16_MAX
#define U32_MAX		UINT32_MAX
#else
typedef unsigned int		a_u32;
typedef unsigned short int	a_u16;
typedef unsigned char		a_u8;
typedef signed int		a_s32;
typedef signed short int	a_s16;
typedef signed char		a_s8;
#define U16_MAX		(0xffffU)
#define U32_MAX		(0xffffffffU)
#endif

#if ! HAVE_SSIZE_T
typedef int ssize_t;
#endif

typedef a_u32 a_keycode; /* FIXME: move to media/ */

#define UCHAR(c) ((unsigned char) (c))

/* xalloc.h (from fileutils.h) define __attribute__ only for gcc >= 2.8
   (according to the ChangeLog this is needed for OPENStep 4.2 cc)
   while we want to allow __attribute__ for gcc >= 2.7 in Heroes
   (because 2.7.2 is still widely used and __attribute__((packed)) is
   mandatory).  So we need to undefine __attribute__ after xalloc has been
   included. */
#ifndef __attribute__
/* remember that __attribute has to be undefined after inclusion */
# define undefine__attribute__
#endif
#include "xalloc.h"
#ifdef undefine__attribute__
# undef __attribute__
# undef undefine__attribute__
#endif

/* Like XFREE, but also zeroes Var */
#define XFREE0(Var)				\
  do {						\
    if (Var) {					\
      free (Var);				\
      (Var) = 0;				\
    }						\
  } while (0)

#define XCALLOC_ARRAY(Array, N_items) \
Array = xcalloc (sizeof (*(Array)), (N_items))

#define XMALLOC_ARRAY(Array, N_items) \
Array = xmalloc (sizeof (*(Array)) * (N_items))

#define XREALLOC_ARRAY(Array, N_items) \
Array = xrealloc ((Array), sizeof (*(Array)) * (N_items))

#define XMALLOC_VAR(Var) XMALLOC_ARRAY ((Var), 1)
#define XCALLOC_VAR(Var) XCALLOC_ARRAY ((Var), 1)

/* Like XMALLOC_ARRAY but also performs a memSet */
#define XSALLOC_ARRAY(Array, N_items, Val)			\
  do {								\
    XMALLOC_ARRAY ((Array), (N_items));				\
    memset ((Array), (Val), sizeof (*(Array)) * (N_items));	\
  } while (0)

#if HAVE_MTRACE
# include <mcheck.h>
#else
# define mtrace()
# define muntrace()
#endif

/* keep this header at the end of the include list, because it may
   define some macros to change the declaration of malloc functions */
#if HAVE_DMALLOC_H
# define DMALLOC_FUNC_CHECK
# include <dmalloc.h>
#endif
#if HAVE_MPATROL_H
# include <mpatrol.h>
#endif

/* miscellaneous prototypes for replacement functions */

#if ! HAVE_DECL_STRCASECMP
int strcasecmp (const char *s1, const char *s2);
#endif

#if ! HAVE_DECL_STPCPY
# ifndef stpcpy
char *stpcpy (const char *s1, const char *s2);
# endif
#endif

#if HAVE_MKDIR
# if MKDIR_TAKES_ONE_ARG
#  define mkdir(a,b) mkdir(a)
# endif
#else
# if HAVE__MKDIR
#  define mkdir(a,b) _mkdir(a)
# else
#  error "Don't know how to create a directory on this system."
# endif
#endif

/* For systems that distinguish between text and binary I/O.
   O_BINARY is usually declared in fcntl.h  */
#if !defined O_BINARY && defined _O_BINARY
  /* For MSC-compatible compilers.  */
# define O_BINARY _O_BINARY
# define O_TEXT _O_TEXT
#endif

#ifdef __BEOS__
  /* BeOS 5 has O_BINARY and O_TEXT, but they have no effect.  */
# undef O_BINARY
# undef O_TEXT
#endif

#ifndef O_BINARY
  /* On reasonable systems, binary I/O is the default.  */
# define O_BINARY 0
# define O_TEXT 0
#endif

/* Define S_ISDIR if it isn't already defined in sys/stat.h */
#if defined(S_IFDIR) && !defined(S_ISDIR)
# define S_ISDIR(mode) (((mode) & S_IFMT) == S_IFDIR)
#endif

/* Re-define S_ISDIR if the native implementation is broken. According to
   the autoconf info-file this is the case for Tektronix UTekV, Amdahl UTS
   and Motorola System V/88 */
#if defined(STAT_MACROS_BROKEN)
# undef S_ISDIR
# define S_ISDIR(mode) (((mode) & S_IFMT) == S_IFDIR)
#endif


#if !S_IRUSR && S_IREAD
# define S_IRUSR S_IREAD
#endif
#if !S_IRUSR
# define S_IRUSR 00400
#endif
#if !S_IRGRP
# define S_IRGRP (S_IRUSR >> 3)
#endif
#if !S_IROTH
# define S_IROTH (S_IRUSR >> 6)
#endif

#if !S_IWUSR && S_IWRITE
# define S_IWUSR S_IWRITE
#endif
#if !S_IWUSR
# define S_IWUSR 00200
#endif
#if !S_IWGRP
# define S_IWGRP (S_IWUSR >> 3)
#endif
#if !S_IWOTH
# define S_IWOTH (S_IWUSR >> 6)
#endif

#if !S_IXUSR && S_IEXEC
# define S_IXUSR S_IEXEC
#endif
#if !S_IXUSR
# define S_IXUSR 00100
#endif
#if !S_IXGRP
# define S_IXGRP (S_IXUSR >> 3)
#endif
#if !S_IXOTH
# define S_IXOTH (S_IXUSR >> 6)
#endif

#if !S_IRWXU
# define S_IRWXU (S_IRUSR | S_IWUSR | S_IXUSR)
#endif
#if !S_IRWXG
# define S_IRWXG (S_IRGRP | S_IWGRP | S_IXGRP)
#endif
#if !S_IRWXO
# define S_IRWXO (S_IROTH | S_IWOTH | S_IXOTH)
#endif

/* S_IXUGO is a common extension to POSIX.1.  */
#if !S_IXUGO
# define S_IXUGO (S_IXUSR | S_IXGRP | S_IXOTH)
#endif

#ifndef S_IRWXUGO
# define S_IRWXUGO (S_IRWXU | S_IRWXG | S_IRWXO)
#endif

#ifndef S_IRUGO
# define S_IRUGO (S_IRUSR | S_IRGRP | S_IROTH)
#endif

#ifndef S_IWUGO
# define S_IWUGO (S_IWUSR | S_IWGRP | S_IWOTH)
#endif

#ifndef S_IRWUGO
# define S_IRWUGO (S_IRUGO | S_IWUGO)
#endif

/* ===== The following ctype definitions are stolen from textutils ===== */

/* Jim Meyering writes:

   "... Some ctype macros are valid only for character codes that
   isascii says are ASCII (SGI's IRIX-4.0.5 is one such system --when
   using /bin/cc or gcc but without giving an ansi option).  So, all
   ctype uses should be through macros like ISPRINT...  If
   STDC_HEADERS is defined, then autoconf has verified that the ctype
   macros don't need to be guarded with references to isascii. ...
   Defining isascii to 1 should let any compiler worth its salt
   eliminate the && through constant folding."

   Bruno Haible adds:

   "... Furthermore, isupper(c) etc. have an undefined result if c is
   outside the range -1 <= c <= 255. One is tempted to write isupper(c)
   with c being of type `char', but this is wrong if c is an 8-bit
   character >= 128 which gets sign-extended to a negative value.
   The macro ISUPPER protects against this as well."  */

#if STDC_HEADERS || (!defined (isascii) && !HAVE_ISASCII)
# define IN_CTYPE_DOMAIN(c) 1
#else
# define IN_CTYPE_DOMAIN(c) isascii(c)
#endif

#ifdef isblank
# define ISBLANK(c) (IN_CTYPE_DOMAIN (c) && isblank (c))
#else
# define ISBLANK(c) ((c) == ' ' || (c) == '\t')
#endif
#ifdef isgraph
# define ISGRAPH(c) (IN_CTYPE_DOMAIN (c) && isgraph (c))
#else
# define ISGRAPH(c) (IN_CTYPE_DOMAIN (c) && isprint (c) && !isspace (c))
#endif

/* This is defined in <sys/euc.h> on at least Solaris2.6 systems.  */
#undef ISPRINT

#define ISPRINT(c) (IN_CTYPE_DOMAIN (c) && isprint (c))
#define ISALNUM(c) (IN_CTYPE_DOMAIN (c) && isalnum (c))
#define ISALPHA(c) (IN_CTYPE_DOMAIN (c) && isalpha (c))
#define ISCNTRL(c) (IN_CTYPE_DOMAIN (c) && iscntrl (c))
#define ISLOWER(c) (IN_CTYPE_DOMAIN (c) && islower (c))
#define ISPUNCT(c) (IN_CTYPE_DOMAIN (c) && ispunct (c))
#define ISSPACE(c) (IN_CTYPE_DOMAIN (c) && isspace (c))
#define ISUPPER(c) (IN_CTYPE_DOMAIN (c) && isupper (c))
#define ISXDIGIT(c) (IN_CTYPE_DOMAIN (c) && isxdigit (c))
#define ISDIGIT_LOCALE(c) (IN_CTYPE_DOMAIN (c) && isdigit (c))

#if STDC_HEADERS
# define TOLOWER(Ch) tolower (Ch)
# define TOUPPER(Ch) toupper (Ch)
#else
# define TOLOWER(Ch) (ISUPPER (Ch) ? tolower (Ch) : (Ch))
# define TOUPPER(Ch) (ISLOWER (Ch) ? toupper (Ch) : (Ch))
#endif

/* ISDIGIT differs from ISDIGIT_LOCALE, as follows:
   - Its arg may be any int or unsigned int; it need not be an unsigned char.
   - It's guaranteed to evaluate its argument exactly once.
   - It's typically faster.
   Posix 1003.2-1992 section 2.5.2.1 page 50 lines 1556-1558 says that
   only '0' through '9' are digits.  Prefer ISDIGIT to ISDIGIT_LOCALE unless
   it's important to use the locale's definition of `digit' even when the
   host does not conform to Posix.  */
#define ISDIGIT(c) ((unsigned) (c) - '0' <= 9)

/* ===== end of ctype definitions ===== */

#if ! HAVE_SETGID
# if HAVE_SETREGID
#  define setuid(X) setreuid(X,X)
#  define setgid(X) setregid(X,X)
# else
#  define setuid(X)
#  define setgid(X)
# endif
#endif

#if ! HAVE_SETEGID
# if HAVE_SETREGID
#  define seteuid(X) setreuid(-1,X)
#  define setegid(X) setregid(-1,X)
# else
#  define seteuid(x)
#  define setegid(x)
# endif
#endif

#if ! HAVE_SETREGID
# define setreuid(x,y)
# define setregid(x,y)
#endif

#if ! HAVE_GETEGID
# define geteuid() 0
# define getuid()  0
# define getegid() 0
# define getgid()  0
#endif


#ifndef __attribute__
# if __GNUC__ < 2 || (__GNUC__ == 2 && __GNUC_MINOR__ < 7) || __STRICT_ANSI__
#  define __attribute__(x)
# endif
#endif

#ifndef ATTRIBUTE_NORETURN
# define ATTRIBUTE_NORETURN __attribute__ ((noreturn))
#endif

#ifndef ATTRIBUTE_UNUSED
# define ATTRIBUTE_UNUSED __attribute__ ((unused))
#endif

#ifndef ATTRIBUTE_PACKED
# define ATTRIBUTE_PACKED __attribute__ ((packed))
#endif

#ifndef ATTRIBUTE_PRINTF
# define ATTRIBUTE_PRINTF(str,fst) __attribute__ ((format (printf, str, fst)))
#endif


#ifndef PTR_TO_INT
# define PTR_TO_INT(P) (((char *) P) - (char *) 0)
#endif
#ifndef INT_TO_PTR
# define INT_TO_PTR(P, TYPE) ((P) + ((TYPE) *) 0)
#endif

/*****************************************************************
******************************************************************
*** extracted from ./lvl/lvl.h
**/

/* Tiles are units of 24x20 pixels, squares are units of 12x10 pixels.
   There is 4 squares per tiles.

   Functions can use squares or tiles, depending on the resolution
   with which entities are stored.  For instance the bonuses are
   stored in a tile-sized array, while the vehicules are stored in an
   squares-sized array.

   Those types are defined for better readability.  The a_*_coord
   types are used to designate a coordinate (either X, or Y), while
   the a_*_index are used for square numbers (indices in one
   dimensional arrays: index = y_coord * width + x_coord).  */
typedef unsigned int a_square_coord;
typedef unsigned int a_square_index;
typedef unsigned int a_tile_coord;
typedef unsigned int a_tile_index;
#define INVALID_INDEX (~0u)

typedef struct a_square_coord_pair a_square_coord_pair;
struct a_square_coord_pair {
  a_square_coord y, x;
};

typedef struct a_tile_coord_pair a_tile_coord_pair;
struct a_tile_coord_pair {
  a_tile_coord y, x;
};

/* Direction constants.  */
typedef enum a_dir a_dir;
enum a_dir { D_UP = 0,
	     D_RIGHT = 1,
	     D_DOWN = 2,
	     D_LEFT = 3 };
typedef a_u8 a_dir8;		/* Hold the same values, on 8bits. */
#define REVERSE_DIR(d) ((d) ^ 2)

/* Two direction on one byte (used for trails).
   Don't change the encoding, this the resulting value is used
   to lookup sprites to draw the trail.  */
typedef a_u8 a_dir8_pair;
#define DIR8_TAIL(X) ((X) >> 2)
#define DIR8_HEAD(X) ((X) & 0x3)
#define DIR8_PAIR(H, T) (((T) << 2) | (H))


/* Number of directions,
   unless Heroes goes 3D this is unlikely to change :) */
#define DIR_MAX 4

/* Similar constants, but OR-able, to indicate multiple directions.  */
typedef enum a_dir_mask a_dir_mask;
enum a_dir_mask { DM_UP = 1,
		  DM_RIGHT = 2,
		  DM_DOWN = 4,
		  DM_LEFT = 8 };
typedef a_u8 a_dir_mask8;	/* Hold the same values, on 8bits. */
#define DM_ALL (DM_UP | DM_RIGHT | DM_DOWN | DM_LEFT)
#define DIR_TO_DIRMASK(dir) (1 << (dir))

/* Tile type constants.  */
typedef enum a_tile_type a_tile_type;
enum a_tile_type { T_NONE = 0,
		   T_STOP = 1,
		   T_SPEED = 2,
		   T_TUNNEL = 3,
		   T_BOOM = 4,
		   T_ANIM = 5,
		   T_ICE = 6,
		   T_DUST = 7,
		   T_OUTWAY = 8,
		   T_MAXTYPE	/* Keep this upper bound last in the enum. */};
typedef a_u8 a_tile_type8;	/* Hold the same values, on 8bits. */


typedef struct a_level a_level;	/* An Heroes level.  */
typedef struct a_level_bits a_level_bits; /* Hiden details about levels.  */
typedef struct a_tileset a_tileset; /* defined below */

/* The pointers in the a_level structure should point to constant
   data.  But from the level loading code the data need to be mutable.
   So we allow const to be removed by defining LVL_MUTABLE.  */

#ifndef LVL_MUTABLE
# define LVL_MUTABLE const
#endif

struct a_level {
  /*-------------.
  | Dimensions.  |
  `-------------*/

  /* We store the dimensions in those two units, so that they are
     directly available.  */
  a_tile_coord tile_height;
  a_tile_coord tile_width;
  a_square_coord square_height;
  a_square_coord square_width;

  /* Some levels are wrapped: when the vehicle reach the right edge of
     the level, they continue as if they were entering the level from
     the left edge, and vice-versa.  Of course this can also be set
     for the top and bottom.

     The _wrap members should be ANDed with a coordinate to wrap it.
     For instance a wrapped level with a width of 32 will have
     `tile_width_wrap' set to 31.  That way an y coordinate of 35 will
     be mapped to 3 (== 35 & 31).  Non-wrapped level set this value to
     `~0u' (== DONT_WRAP) so that ANDing is no-op.  */
  a_tile_coord tile_height_wrap;
  a_tile_coord tile_width_wrap;
  a_square_coord square_height_wrap;
  a_square_coord square_width_wrap;
#ifndef DONT_WRAP
# define DONT_WRAP (~0u)	/* Used for non wrapping directions.  */
#endif

  /* Most maps are allocated as one-dimensional arrays.  So we define
     two more constants used for the size of the arrays.  */
  a_tile_index tile_count;	/* == tile_width * tile_height */
  a_square_index square_count;	/* == square_width * square_height */

  /*----------------------------------------------------------------.
  | The following arrays are initialized only if the level body has |
  | been loaded.  See the load_body argument of lvl_load_file.      |
  `----------------------------------------------------------------*/

  /* Type of each square.  */
  LVL_MUTABLE a_tile_type8 *square_type;
  /* Set of blocked directions for each square.  If the entry is set
     to DM_UP|DM_LEFT, it means a vehicle cannot *leave* the square
     by the top or left edges.  */
  LVL_MUTABLE a_dir_mask8 *square_walls_out;
  /* Directions for squares for which it matters (i.e., T_SPEED, and
     T_TUNNEL).  */
  LVL_MUTABLE a_dir8 *square_direction;
  /* square_move[D_UP][SQR] is the square that a vehicle shoud go into
     when it leave SQR by the top edge.  */
  LVL_MUTABLE a_square_index *square_move[DIR_MAX];

  /* Optionally loaded tileset information -- only needed for map editing */
  LVL_MUTABLE a_tileset *tileset;

  /*-------------------------------------------------------------.
  | Opaque structure to hide the remaining data.  Use one of the |
  | following functions to access anything relevant in there.    |
  `-------------------------------------------------------------*/

  LVL_MUTABLE a_level_bits *private;
};

/*****************************************************************
******************************************************************
*** extracted from ./lvl/lvl.h
**/

/* These two functions return a pointer to a string allocated for LVL,
   and freed by lvl_free().  You should not free them yourself, and
   you should not use these results after a call to lvl_free().  */
const char *lvl_sound_track (const a_level *lvl);
const char *lvl_tile_sprite_map_basename (const a_level *lvl);

/* Initialize COORD, and DIR, with the starting position for PLAYER.  */
void lvl_start_position (const a_level *lvl, unsigned int player,
			 a_square_coord_pair *coord, a_dir *dir);

/*--------------------------------------------------------------------.
| NOTE: The following function will fail if the body of the level     |
| has not been loaded (see the load_body argument of lvl_load_file).  |
`--------------------------------------------------------------------*/

/* Return the type of tile TILE on level LVL.  */
a_tile_type lvl_tile_type (const a_level *lvl, a_tile_index tile);

/* Return the output direction of a tunnel on tile TILE.  This function
   should only be called on tiles of type T_TUNNEL.  */
a_dir lvl_tunnel_output_dir (const a_level *lvl, a_tile_index tile);

/* Return the offset (to the first pixel) of the sprite used to draw TILE.  */
unsigned int lvl_tile_sprite_offset (const a_level *lvl,
				     a_tile_index tile);
/* Return the offset (to the first pixel) of the overlay sprite to
   draw on TILE.  */
unsigned int lvl_tile_sprite_overlay_offset (const a_level *lvl,
					     a_tile_index tile);

/* Kind of animation.  */
typedef enum an_anim_kind an_anim_kind;
enum an_anim_kind { A_NONE, A_LOOP, A_PINGPONG };

/* Return information about a possible animation on tile TILE.
   FRAME_COUNT is the number of frames to display, and DELAY a number
   of 1/70sec to wait between each frame.  Actually, FRAME_COUNT is
   the number of sprites to display: FRAME_COUNT sprites starting at
   the sprite index for the current tile should be displayed.

   There is no animation if FRAME_COUNT is set to 0.  */
void lvl_animation_info (const a_level *lvl, a_tile_index tile,
			 unsigned int *frame_count, unsigned int *delay,
			 an_anim_kind *kind);

/*----------------.
| Useful macros.  |
`----------------*/

/* Moving to neighbor squares.  */
#define SQR_COORD_LEFT(lvl_ptr, sqr_x_coord) \
  (((sqr_x_coord) - 1) & (lvl_ptr)->square_width_wrap)
#define SQR_COORD_RIGHT(lvl_ptr, sqr_x_coord) \
  (((sqr_x_coord) + 1) & (lvl_ptr)->square_width_wrap)
#define SQR_COORD_UP(lvl_ptr, sqr_y_coord) \
  (((sqr_y_coord) - 1) & (lvl_ptr)->square_height_wrap)
#define SQR_COORD_DOWN(lvl_ptr, sqr_y_coord) \
  (((sqr_y_coord) + 1) & (lvl_ptr)->square_height_wrap)

/* Checking square coordinates.  */
#define SQR_COORD_X_VALID(lvl_ptr, sqr_coord_x) \
  (((sqr_coord_x) < (lvl_ptr)->square_width))
#define SQR_COORD_Y_VALID(lvl_ptr, sqr_coord_y) \
  (((sqr_coord_y) < (lvl_ptr)->square_height))
#define SQR_COORD_VALID(lvl_ptr, sqr_coord_y, sqr_coord_x) \
  (SQR_COORD_Y_VALID (lvl_ptr, sqr_coord_y) \
   && SQR_COORD_X_VALID (lvl_ptr, sqr_coord_x))

/* Index to coordinate conversions.  */
#define SQR_INDEX_TO_COORD_X(lvl_ptr, sqr_index) \
  ((sqr_index) % (lvl_ptr)->square_width)
#define SQR_INDEX_TO_COORD_Y(lvl_ptr, sqr_index) \
  ((sqr_index) / (lvl_ptr)->square_width)
#define SQR_COORDS_TO_INDEX(lvl_ptr, sqr_y, sqr_x) \
  ((lvl_ptr)->square_width * (sqr_y) + (sqr_x))

#define TILE_INDEX_TO_COORD_X(lvl_ptr, tile_index) \
  ((tile_index) % (lvl_ptr)->tile_width)
#define TILE_INDEX_TO_COORD_Y(lvl_ptr, tile_index) \
  ((tile_index) / (lvl_ptr)->tile_width)
#define TILE_COORDS_TO_INDEX(lvl_ptr, tile_y, tile_x) \
  ((lvl_ptr)->tile_width * (tile_y) + (tile_x))


/* Tile indexes to square indexes conversions.  */
#define TILE_INDEX_TO_SQR_INDEX(lvl_ptr, tile_index)		\
  (TILE_INDEX_TO_COORD_Y ((lvl_ptr), (tile_index)) * 2		\
   * (lvl_ptr)->square_width					\
   + TILE_INDEX_TO_COORD_X ((lvl_ptr), (tile_index)) * 2)

#define SQR_INDEX_TO_TILE_INDEX(lvl_ptr, sqr_index)		\
  ((SQR_INDEX_TO_COORD_Y ((lvl_ptr), (sqr_index)) / 2)		\
   * (lvl_ptr)->tile_width					\
   + (SQR_INDEX_TO_COORD_X ((lvl_ptr), (sqr_index)) / 2))

/* Given the index of the top left square of a tile,
   compute the index for each square in the tile.  */
#define SQR0(lvl_ptr, sqr0) (sqr0)
#define SQR1(lvl_ptr, sqr0) ((sqr0) + 1)
#define SQR2(lvl_ptr, sqr0) ((sqr0) + (lvl_ptr)->square_width)
#define SQR3(lvl_ptr, sqr0) ((sqr0) + (lvl_ptr)->square_width + 1)
#define SQRX(lvl_ptr, sqr0, x) \
  ((sqr0) + (((x) & 2) ? (lvl_ptr)->square_width : 0) + ((x) & 1))

/*****************************************************************
******************************************************************
*** extracted from ./rules/bonuses.h
**/

typedef enum a_bonus a_bonus;
typedef a_u8 a_bonus8;		/* Usually, we store bonus on 8bits. */

enum a_bonus {
  B_NOTHING,

  B_LENGTH_UP,			/* L+ */
  B_LENGTH_DOWN,		/* L- */
  B_SPEED_UP,			/* S+ */
  B_SPEED_DOWN,			/* S- */
  B_RANDOM,			/* ? */
  B_POINTS,			/* C (probably this stood for `Coins') */
  B_FIRE_TRAIL,			/* ZZ */
  B_PAUSE,			/* !! */
  B_INVERTED,			/* -1 */
  B_TURBO_UP,			/* T+ */
  B_TURBO_DOWN,			/* T- */
  B_END_OF_LEVEL,		/* EL */
  B_INVINCIBLE,			/* [] */
  B_ROTOZOOM,			/* X */
  B_EXTRA_LIFE,			/* XL */
  B_WAVES,			/* ~~ */
  B__UNUSED_YET,

  /* This one is not drawn in bonus{a,b}.pcx because it's not really
     bonus, but we handle it like the other bonus.  */
  B_BIG_DOLLAR,

  /* `|' this one with the above values to yellow bonuses instead of
     purple bonuses.  */
  B_YELLOW = 128,

  /* Indicate that this place should never contain a bonus.  */
  B_NOT_HERE = 255,
};

#define BONUS_TYPE(b)  ((b) & 127)
#define BONUS_YELLOW_P(b) (((b) & B_YELLOW) == B_YELLOW)
#define BONUS_P(b) ((b) != B_NOTHING && (b) != B_NOT_HERE)

/*****************************************************************
******************************************************************
*** extracted from ./rules/player.h
**/

/* these two structures are used to access
   the higher and lower part of a a_u32 */
typedef struct {
#ifdef WORDS_BIGENDIAN
  a_u16 h, l;
#else
  a_u16 l, h;
#endif
} hl;

typedef union {
  a_u32 e;
  hl h;
} ehl;

/* FIXME: There are meny fields here that should not be public.  */

typedef struct a_player a_player;
struct a_player
{
  /* Position, in various format.  */
  a_tile_coord tx, ty;
  a_square_coord sx, sy;
  a_square_index si;

  int vitp;			/* current speed */
  ehl d;			/* offset */

  a_dir way;			/* Actual heading direction.  */
  a_dir old_way;		/* Previous one (i.e. was the vehicle
				   was on the previous square; it may be the
				   same.  */
  a_dir tunnel_way;		/* After running throught a tunnel, the
				   vehicle might get out with another heading
				   direction (tunnel_way) than on entrance
				   (way).  */
  a_dir next_way;		/* The direction that shall be taken on next
				   square, if possible.  */

  int turbo;			/* Throttle. */

  int inversed_controls;	/* inverted commands */

  unsigned int score;		/* The present score.  */
  unsigned int score_delta;	/* A value stepping towards "score", for
				   the renderer (FIXME: this ought to
				   belong to the renderer data).  */

  int turbo_level;		/* The present value for the turbo gauge. */
  int turbo_level_delta;	/* A value stepping toward "turbo_level,
				   for the renderer (FIXME: this ought
				   to belong to the renderer data).  */

  int delay;			/* delay frames (FIXME: don't use frames) */
  int speedup;			/* bonus speedup or speeddown */
  int rotozoom;			/* roto */
  int rotozoom_direction;
  int waves;
  int waves_begin;
  int fire_trail;
  int invincible;		/* blinking */

  int lifes;			/* lifes LEFT */
  char notify_delay;		/* a pause is coming */
  char tunnel_inverse;
  char autopilot;
  char cpu;			/* 0: local CPU    [1: remote CPU] */
				/* 2: player local [,3: player distant] */
  int lemmings_nbr;
  int martians_nbr;		/* ;-) */
  int time;
  int cash;			/* ... or colors */
  int wins;			/* games won so far */

  /* FIXME: Get rid of these horrors.  */
  int spec;			/* special event (tunnel,ice,death) */
  int ai_max_depth;		/* recusrion depth for CPU (keep <= 7,
				   or it will be slow) */
  int target;			/* target to follow */

};

/*****************************************************************
******************************************************************
*** extracted from ./rules/state.h
**/

/* Information about a lemming.  */
typedef struct a_lemming a_lemming;
struct a_lemming {
  /* Position of the lemming.

     Lemmings are always astride on two squares when moving.  If
     pos_head == pos_tail, then the lemmings is stopped.

     When the lemming is moving, the offset within the current square
     is common with all lemmings (they move synchornously) and
     returned by the state_lemmings_move_offset() function.  */
  a_square_index pos_head, pos_tail;

  /* Direction the lemming looks towards.  */
  a_dir dir;

  /* Color of the lemmings.  */
  int color;

  /* True iff the lemming is dead.  */
  bool dead;

  /* For dead lemmings, i.e. blood stains, the position within the
     square is NOT the one given by state_lemmings_move_offset(),
     otherwise the stains would be moving... */
  unsigned int puddle_offset;

  /* Several puddle can be accumulated on the same square.  */
  a_lemming *next_puddle;
};

/* When playing "Kill'em All", each player start with
   LEMMINGS_PER_PLAYERS lemmings.  */
#define LEMMINGS_PER_PLAYERS 50

/* The game modes, as selected in the PLAY menu.  */
typedef enum a_game_mode a_game_mode;
enum a_game_mode {
  M_QUEST = 0,			/* Quest */
  M_DEATHM = 1,			/* Death-match */
  M_KILLEM = 2,			/* Kill'em all */
  M_TCASH = 3,			/* Time-ca$h */
  M_COLOR = 4,			/* Color */
};

/* An explosions is displayed using NBR_EXPLOSION_FRAMES intermediate
   sprites.  There is NBR_EXPLOSION_KINDS different set of explosions
   sprites (the set is chosen randomly before an explosion is displayed,
   mixing them make large flames looks better).
   It should be easy the add new kinds of explosions.  Adding more
   frame however would need several timing adjustement in the code.  */
#define NBR_EXPLOSION_FRAMES 15
#define NBR_EXPLOSION_KINDS   2

/* Number of time slices for each frames.  Changing this affects
   the speed of the explosions.  */
#define EXPLOSION_SLICES_PER_FRAMES 8
/* Number of frames between the moment an explosions is triggered, and
   the moment it actually explodes.  */
#define EXPLOSION_DELAY 8

/*
 *  Various constants, used for the an_explosion type.
 */

/* The explosion will explode immediately.  */
#define EXPLOSION_IMMEDIATE (NBR_EXPLOSION_FRAMES - 1)
/* The explosion is triggered, it will explode later.  */
#define EXPLOSION_TRIGGERED (EXPLOSION_IMMEDIATE + EXPLOSION_DELAY)
/* It's time for the explosion to trigger neighbors.  */
#define EXPLOSION_TRIGGER_NEIGHBORS (EXPLOSION_TRIGGERED - 3)
/* The explosion is not triggered (it's also triggerable).  */
#define EXPLOSION_UNTRIGGERED (EXPLOSION_TRIGGERED + 1)

/*
 * 0 <= n < NBR_EXPLOSION_FRAMES: frame number to display.
 * NBR_EXPLOSION_FRAMES <= n <= EXPLOSION_TRIGGERED: about to explode.
 * n == EXPLOSION_UNTRIGGERED: idle.
 */
typedef a_u8 an_explosion;
#define EXPLOSION_SQUARE_TRIGGERED_P(state_ptr, idx)		\
  ((state_ptr)->square_explo_state[idx] <= EXPLOSION_TRIGGERED)
#define EXPLOSION_SQUARE_TRIGGERABLE_P(state_ptr, idx)		\
  ((state_ptr)->level->square_type[idx] == T_BOOM		\
   && (state_ptr)->square_explo_state[idx] == EXPLOSION_UNTRIGGERED)

/* Some pointers in the a_level_state structure should point to
   constant data, so the user a warned if s/he modify it.  However we
   want to allow the functions from the libhrules.a library to modify
   them. So we allow const to be removed by defining
   LVL_STATE_MUTABLE.  */

#ifndef LVL_STATE_MUTABLE
# define LVL_STATE_MUTABLE const
#endif

/* Dynamic state of a level.  */
typedef struct a_level_state a_level_state;
/* Hidden details in a_level_state */
typedef struct a_level_state_bits a_level_state_bits;

struct a_level_state {
  /* The level definition. (Always const.) */
  const a_level *level;

  /* Informations about each player.  */
  a_player *player[4];		/* FIXME: use LVL_STATE_MUTABLE */

  /* State of each square.
     0xFF: free
     4,5,6,7: heads of the vehicles
     0,1,2,3: tails of the vehicles (and heads of the trails)
     8,9,10,11: trails
     12,13,14,15: tails of the trails.
  */
  LVL_STATE_MUTABLE a_u8 *square_occupied;
#define SQOC_VEHICLE_TAIL(color) (color)
#define SQOC_VEHICLE_HEAD(color) ((color) + 4)
#define SQOC_TRAIL(color) ((color) + 8)
#define SQOC_TRAIL_TAIL(color) ((color) + 12)
#define SQOC_VACANT 0xFFU
  /* SQOC = SQuare OCcupied.  */

  /* FIXME: Document or remove.  */
  int col2plr[4];
  int plr2col[4];

  /* If the square is occupied, this indicate the directions of the
     trail.  */
  LVL_STATE_MUTABLE a_dir8_pair *square_way;

  /* Fast conversion between a_square_index and a_tile_index.  */
  LVL_STATE_MUTABLE a_tile_index *square_tile;

  /* Fast conversion between a_square_index and a_square_coord_pair.  */
  a_square_coord_pair *square_coord;

  /* Objet lying on the level (pyramids, dollars, clocks, ...).
      -1: nothing
      -2: can't drive here
     In TIME CASH mode:
      0: dollar
      15: clock
     In COLOR mode:
      0,1,2,3,4: color gems (4 = grey)
      8,9,10,11,12: checked color gems
      16: clock
      24: checked clock
   */
  a_s8 *square_object;
#define SQOB_NOTHING (-1)
#define SQOB_UNREACHABLE (-2)
  /* SQOB = SQuare OBject.  */

  a_game_mode game_mode;

  /* The state of the explosion for each square.  See the EXPLOSION_*
     constants above.  */
  LVL_STATE_MUTABLE an_explosion *square_explo_state;
  /* The type of each explosion, smaller than NBR_EXPLOSION_KINDS
     for each square.   This is really usefull only to render
     the level; all explosions have the same effect on the players. */
  LVL_STATE_MUTABLE a_u8 *square_explo_type;

  /* For each tile, the number of the bonus present, if any.  */
  LVL_STATE_MUTABLE a_bonus8 *tile_bonus;

  /* For each square occupied by a lemming, this holds a pointer to
     the lemming data.  If the square is empty, the pointer is NULL.  */
  a_lemming **square_lemmings_list;

  /* For each square occupied by a puddle, this holds a pointer to the
     lemmings data.  Several puddles might be chained with the
     next_puddle fieds of this structure.  The pointer is NULL if
     there is no puddle.  Blood puddles have no effect on the players
     (as far the rules of the game are concerned), so this array
     should be useful only to the renderer.  */
  a_lemming **square_dead_lemmings_list;

  /* Private data.  Use the state_* functions to access them.  */
  LVL_STATE_MUTABLE a_level_state_bits *private;
};

/* Size of the player's trail (i.e. number of L+ bonus eaten + 1).  */
int state_trail_size (const a_level_state *state, int player);

/* Return TRUE if the trail is expending.  */
bool state_trail_expending (const a_level_state *state, int player);

/*****************************************************************
******************************************************************
*** extracted from ./hooks.h
**/

/* Plug-ins can "hook" some functions which must be called back
   whenever a "hook point" is encountered.

   This implementation is heavily inspired from the one found in Zsh.
*/

typedef struct a_hook a_hook;
typedef struct a_hook_fun_list a_hook_fun_list;
typedef void (*a_hook_fun)(const a_hook *hook,
			   void *hook_data,
			   void *callback_data);

struct a_hook {
  a_hook *next;
  const char *name;
  a_hook_fun_list *list;
};

#define HOOK_DEF(name)  { 0, name, 0 }

/* The code which provide a hook should define a hook structure as
   follow

   static a_hook sample_hook = HOOK_DEF ("sample");

   and call hook_define to make this hook available to this other part
   of the program:

   hook_define (&sample_hook);

   Whenevert the point where this hook should be run is encountered,
   the hook_run function should be used:

   hook_run (&sample_hook, 0);

   The second argument of hook_run may hold additional data which will
   be passed to all functions run for this hook.  (It is up to the
   provider and user of a hook to use the right convention to access
   the supplied data).  */
void hook_define (a_hook *hook);
void hook_run (a_hook *hook, void *hook_data);

/* Same a hook_define, but works on an array of hooks.  */
void hook_define_many (a_hook *hook, int size);

void hook_undefine (a_hook *hook);
void hook_undefine_many (a_hook *hook, int size);


/* One can subscribe to a hook using the hook_add_fun function, and
   unsubscribe using the hook_rem_fun function.

   The functions added to a hook should have the following prototype:

   void
   sample_hook_fun (const a_hook *hook, void *hook_data, void *callback_data)
   {
     ...
   }

   The HOOK pointer points to the hook for which is function is being
   called back.  This is almost useless, except to get the name
   of this hook (hook->name).

   HOOK_DATA might point to additional data provided by the hook provider,
   its interpretation is hook-dependant.

   CALLBACK_DATA is a pointer that was supplied when the function was
   added registered for the hook being run. */
void hook_add_fun (const char *name, a_hook_fun fun, void *callback_data);

/* This remove the _first_ occurence of fun from the list of functions
   registered for a hook.  The returned pointer is the CALLBACK_DATA
   supplied to hook_add_fun.  */
void *hook_rem_fun (const char *name, a_hook_fun fun);

/* Remove an occurence of FUN from the list of functions registered
   for a hook.  The occurence removed is the first occurence for which
   IF_FUN returns true.

   This is usually used to remove a function whose CALLBACK_DATA (as
   indicated to hook_add_fun) matches something indicated by REF_DATA
   (passed to hook_rem_fun_if).  */

typedef bool (*a_hook_rem_if_fun)(void *ref_data, void *callback_data);
void *hook_rem_fun_if (const char *name, a_hook_fun fun, void *ref_data,
		       a_hook_rem_if_fun if_fun);

/*****************************************************************
******************************************************************
*** extracted from ./opponents.h
**/

typedef enum an_opponent_type an_opponent_type;
enum an_opponent_type {
  OT_NONE   = 0,		/* cannot play */
  OT_QUEST  = 1,		/* can play quest mode */
  OT_DEATHM = 2,		/* can play death-math */
  OT_KILLEM = 4,		/* can play kill'em all */
  OT_TCASH  = 8,		/* can play time-ca$h */
  OT_COLOR  = 16,		/* can play color */
};

typedef enum a_throttle a_throttle;
enum a_throttle {
  TH_BRAKE = 0,
  TH_NORMAL = 1,
  TH_SPEEDUP = 2,
};


typedef struct an_opponent_action an_opponent_action;
struct an_opponent_action {
  a_dir dir;
  a_throttle throttle;
};

typedef struct an_opponent_sig an_opponent_sig;
struct an_opponent_sig {

  /* A name for the opponent plug-in.  This is not yet used today,
     except for debugging messages, bug it might be used later to
     present to the user a list of plug-ins to chose from.  */
  const char *name;

  /* A mask indicating which mode this opponent can be used for.  */
  an_opponent_type type;

  /* This is called after the state have been initialized, so that the
     opponent can prepare itself to play.  This function is called
     once for each player that will be driven by the current opponent
     module.  The returned void* data, are callback data which will be
     passed back to the opponent in following function calls.  */
  void *(*initialize_player)(const a_level_state *state, int player);

  /* This is called after the level is over, so the opponent can
     do some cleanup.  For instance callback_data should be
     free here if it was allocated by initialize_player.  */
  void (*finalize_player)(const a_level_state *state, int player,
			  void *callback_data);

  /* Unless NULL, this function is called every time the player is
     introduced into the game (after a death, or at the beginning of
     the game).  The returned void* data will replace any previous
     callback data setup by initialize_player, in following function
     calls.  */
  void *(*start_player)(const a_level_state *state, int player,
			void *callback_data);

  /* Unless NULL, this function is called every time the player is
     removed from the game (i.e., after a death).  */
  void *(*stop_player)(const a_level_state *state, int player,
		       void *callback_data);


  /* Unless NULL, this function is called each time the player enter a
     new square.  This is usually where the player should decide
     what's the next direction it will take.  */
  void (*square_update)(const a_level_state *state, int player,
			an_opponent_action *action,
			void *callback_data);

  /* Unless NULL, this function is called on each frame (i.e., 70
     times per second).  This is called frequently, so please keep
     it fast!  */
  void (*frame_update)(const a_level_state *state, int player,
		       an_opponent_action *action,
		       void *callback_data);
};

void opponent_register (an_opponent_sig *opp);
void opponent_unregister (an_opponent_sig *opp);

#endif /* HEROES__PLUGIN__H */
